/* global jake:true, desc:true, task:true, complete:true, require:true, console:true, process:true */
/* jshint unused:false */
var fs = require('fs');
var path = require('path');
var exec = require('child_process').exec;

var tools = require('./src/moxie/build/tools');
var utils = require('./src/moxie/build/utils');
var wiki = require('./src/moxie/build/wiki');
var mkjs = require('./src/moxie/build/mkjs');

var copyright = [
	"/**",
	" * Plupload - multi-runtime File Uploader",
	" * v@@version@@",
	" *",
	" * Copyright @@year@@, Ephox",
	" * Released under AGPLv3 License.",
	" *",
	" * License: http://www.plupload.com/license",
	" * Contributing: http://www.plupload.com/contributing",
	" *",
	" * Date: @@releasedate@@",
	" */"
].join("\n");


desc("Default build task");
task("default", ["moxie", "mkjs", "docs"], function (params) {});



desc("Build release package");
task("release", ["default", "package"], function (params) {});



desc("Build mOxie");
task("moxie", [], function (params) {
	var moxieDir = "./src/moxie";
	var currentDir = process.cwd();

	process.chdir(moxieDir);
	exec("jake", function(error, stdout, stderr) {
		if (error) {
			console.info(error);
		}
		process.chdir(currentDir);
		complete();
	});
}, true);


desc("Build release files");
task("mkjs", [], function () {
	var amdlc = require('amdlc');
	var uglify = tools.uglify;
	var targetDir = "js";
	var flags = utils.asFlags(arguments);

	var uglifyOptions = {
		unused: true,
		dead_code: true,
		global_defs: {
			MXI_DEBUG: false
		}
	};

	var options = {
		from: ['Uploader.js'],
		compress: uglifyOptions,
		baseDir: 'src',
		rootNS: "plupload",
		expose: "all",
		libs: {
			'moxie': {
				rootNS: 'moxie',
				baseDir: 'src/moxie/src/javascript',
				expose: flags['hide-moxie'] ? false : 'public',
				force: true
			}
		},
		verbose: true,
		outputSource: false,
		outputMinified: false,
		outputDev: targetDir + "/plupload.dev.js",
		outputCoverage: targetDir + "/plupload.cov.js"
	};

	if (!flags['no-image']) {
		options.from.push('ImageResizer.js');
	}


	[].push.apply(options.from,
		mkjs.getExtensionPaths({
			baseDir: options.baseDir,
			extensionsDir: options.libs.moxie.baseDir,
			imageSupport: !flags['no-image']
		})
	);

	// start fresh
	if (fs.existsSync(targetDir)) {
		jake.rmRf(targetDir);
	}
	jake.mkdirP(targetDir);

	amdlc.compile(options);

	// Compile source
	amdlc.compile(utils.extend({}, options, {
		expose: 'public',
		outputSource: targetDir + "/plupload.js",
		outputMinified: false,
		outputDev: false,
		outputCoverage: false
	}));

	// Include Plupload source
	var sourceCode = fs.readFileSync(targetDir +  '/plupload.js').toString();
	if (process.env.umd != 'no') {
		fs.writeFileSync(targetDir + '/plupload.js', mkjs.addUMD("plupload", sourceCode));
	}

	// Minify Plupload and combine with mOxie
	console.info("Writing minified version output to: " + targetDir + "/plupload.min.js");
	uglify(targetDir + '/plupload.js', targetDir + "/plupload.min.js", uglifyOptions);

	// add debug constant to dev source
	mkjs.addDebug(targetDir + "/plupload.dev.js");
	mkjs.addDebug(targetDir + "/plupload.cov.js");
	mkjs.addDebug(targetDir + "/plupload.js");


	// Copy compiled moxie files
	if (flags['no-image']) {
		tools.copySync("src/moxie/bin/flash/Moxie.min.swf", "js/Moxie.swf");
		tools.copySync("src/moxie/bin/silverlight/Moxie.min.xap", "js/Moxie.xap");
	} else {
		tools.copySync("src/moxie/bin/flash/Moxie.swf", "js/Moxie.swf");
		tools.copySync("src/moxie/bin/silverlight/Moxie.xap", "js/Moxie.xap");
	}

	var info = require("./package.json");
	info.copyright = copyright;
	tools.addReleaseDetailsTo(targetDir + "/plupload.dev.js", info);
	tools.addReleaseDetailsTo(targetDir + "/plupload.min.js", info);
	tools.addReleaseDetailsTo(targetDir + "/plupload.js", info);

	// Copy UI Plupload
	jake.cpR("./src/jquery.ui.plupload", targetDir + "/jquery.ui.plupload", {});
	uglify(targetDir + "/jquery.ui.plupload/jquery.ui.plupload.js", targetDir + "/jquery.ui.plupload/jquery.ui.plupload.min.js");

	// Copy Queue Plupload
	jake.cpR("./src/jquery.plupload.queue", targetDir + "/jquery.plupload.queue", {});
	uglify(targetDir + "/jquery.plupload.queue/jquery.plupload.queue.js", targetDir + "/jquery.plupload.queue/jquery.plupload.queue.min.js");


	// Add I18n files
	if (flags['i18n']) {
		process.env.auth = "moxieuser:12345";
		process.env.to = "./js/i18n";
		jake.Task['i18n'].invoke();
	}
});



desc("Language tools");
task("i18n", [], function(params) {
	var i18n = require('./build/i18n');

	switch (params) {
		case 'extract':
			var from = process.env.from || ['./src/plupload.js', './src/jquery.ui.plupload/jquery.ui.plupload.js', './src/jquery.plupload.queue/jquery.plupload.queue.js'];
			var to = process.env.to || './tmp/en.po';
			i18n.extract(from, to);
			break;

		case 'toPO':
			// srcLang is required here, e.g.: jake i18n[toPO] srcLang=tmp/en.pot from=js/i18n/*.js
			var from = process.env.from;
			var to = process.env.to || './tmp/i18n';
			i18n.toPot(from, to);
			break;

		case 'pull':
		default:
			var auth = (process.env.auth || "moxieuser:12345").split(':');
			var to = process.env.to || './tmp/i18n';
			i18n.pull(utils.format("https://%s:%s@www.transifex.com/api/2/project/plupload/resource/core/", auth[0], auth[1]), to, complete);
			return;
	}

	complete(); // call complete manually
}, true);



desc("Generate documentation using YUIDoc");
task("docs", [], function (params) {
	var yuidoc = tools.yuidoc;
	yuidoc(["src", "src/jquery.plupload.queue", "src/jquery.ui.plupload"], "docs", {
		norecurse: true
	});
}, true);



desc("Generate wiki pages");
task("wiki", ["docs"], function() {
	wiki("git@github.com:moxiecode/plupload.wiki.git", "wiki", "docs");
});



desc("Runs JSHint on source files");
task("jshint", [], function (params) {
	var jshint = tools.jshint;
	jshint("src", {
		curly: true
	});
});



desc("Package library");
task("package", [], function (params) {
	var zip = tools.zip;
	var info = require("./package.json");

	var tmpDir = "./tmp";
	if (fs.existsSync(tmpDir)) {
		jake.rmRf(tmpDir);
	}
	fs.mkdirSync(tmpDir, 0755);

	var suffix = info.version.replace(/\./g, '_');
	if (/(?:beta|alpha)/.test(suffix)) {
		var dateFormat = require('dateformat');
		// If some public test build, append build number
		suffix += "." + dateFormat(new Date(), "yymmddHHMM", true);
	}

	zip([
		"js/**/*",
		"examples/**/*",
		"README.md",
		"LICENSE.txt"
	], path.join(tmpDir, utils.format("plupload_%s.zip", suffix)), complete);
}, true);











